"""Abstract class for writing chat clients."""

from tornado import gen, httpclient, ioloop
import hashlib
import json
import logging
import random
import re
import time

from hangups import javascript, parsers, exceptions, http_utils, channel, event

logger = logging.getLogger(__name__)
ORIGIN_URL = 'https://talkgadget.google.com'
# Set the connection and request timeouts low so we fail fast when there's a
# network problem.
CONNECT_TIMEOUT = 10
REQUEST_TIMEOUT = 10


def _parse_user_entity(entity):
    """Parse entities returned from the getentitybyid endpoint.

    Raises ValueError if the entity cannot be parsed.
    """
    # Known entity types:
    # GAIA: regular user
    # INVALID: entity does not exist
    entity_type = entity.get('entity_type', None)
    if entity_type != 'GAIA':
        raise ValueError('Cannot parse entity with entity type {}'
                         .format(entity_type))
    try:
        chat_id = entity['id']['chat_id']
        gaia_id = entity['id']['gaia_id']
    except KeyError:
        raise ValueError('Cannot determine entity ID')
    properties = entity.get('properties', {})
    return {
        'chat_id': chat_id,
        'gaia_id': gaia_id,
        'first_name': properties.get('first_name', 'UNKNOWN'),
        'full_name': properties.get('display_name', 'UNKNOWN'),
    }


class ConversationList(object):
    """Wrapper around Client that presents a list of Conversations."""

    def __init__(self, client):
        self._client = client
        self._conv_dict = client.initial_conversations
        logger.info('ConversationList initialized with {} conversation(s)'
                    .format(len(self._conv_dict)))
        # Register event handlers:
        self._client.on_message.add_observer(self._on_message)
        self._client.on_typing.add_observer(self._on_typing)
        self._client.on_focus.add_observer(self._on_focus)
        self._client.on_conversation.add_observer(self._on_conversation)

    def _on_message(self, chat_message):
        """Route on_message event to appropriate Conversation."""
        self.get(chat_message.conv_id).on_message.fire(chat_message)

    def _on_typing(self, typing_message):
        """Route on_typing event to appropriate Conversation."""
        self.get(typing_message.conv_id).on_typing.fire(typing_message)

    def _on_focus(self, focus_message):
        """Route on_focus event to appropriate Conversation."""
        self.get(focus_message.conv_id).on_focus.fire(focus_message)

    def _on_conversation(self, conversation_message):
        """Route on_conversation event to appropriate Conversation."""
        self.get(conversation_message.conv_id).on_conversation.fire(
            conversation_message
        )

    def get_all(self):
        """Return list of all Conversations."""
        return list(self._conv_dict.values())

    def get(self, conv_id):
        """Return a Conversation from its ID.

        Raises KeyError if the conversation ID is invalid.
        """
        return self._conv_dict[conv_id]


class Conversation(object):
    """Wrapper around Client for working with a single chat conversation."""

    def __init__(self, client, id_, users, last_modified, chat_name,
                 chat_messages):
        self._client = client
        self._id = id_ # ConversationID
        self._users = {user.id_: user for user in users} # {UserID: User}
        self._last_modified = last_modified # datetime
        self._name = chat_name # str
        self._chat_messages = chat_messages # ChatMessage

        # Event fired when a new message arrives with arguments (chat_message).
        self.on_message = event.Event('Conversation.on_message')
        # Event fired when a users starts or stops typing with arguments
        # (typing_message).
        self.on_typing = event.Event('Conversation.on_typing')
        # Event fired when a user changes focus on a conversation with
        # arguments (focus_message).
        self.on_focus = event.Event('Conversation.on_focus')
        # Event fired when a conversation updates with arguments
        # (conversation_message).
        self.on_conversation = event.Event('Conversation.on_conversation')


    @property
    def id_(self):
        """Return the Conversation's ID."""
        return self._id

    @property
    def users(self):
        """Return the list of Users participating in the Conversation."""
        return list(self._users.values())

    def get_user(self, user_id):
        """Return a participating use by UserID.

        Raises KeyError if the user ID is not a participant.
        """
        return self._users[user_id]

    @property
    def name(self):
        """ Return chat name if it was renamed manually or None
        :rtype: str
        """
        return self._name

    @property
    def last_modified(self):
        """Return the timestamp of when the conversation was last modified."""
        return self._last_modified

    @property
    def chat_messages(self):
        """Return a list of ChatMessages, sorted oldest to newest."""
        return list(self._chat_messages)

    @gen.coroutine
    def send_message(self, text):
        """Send a message to this conversation.

        text may not be empty.

        Raises hangups.NetworkError if the message can not be sent.
        """
        yield self._client.sendchatmessage(self._id, text)


# TODO: This class isn't really being used yet.
class UserList(object):
    """Allows querying known chat users."""

    def __init__(self, client):
        self._client = client
        self._users = dict(client.initial_users)
        self._self_user_id = client.self_user_id
        logger.info('UserList initialized with {} user(s)'
                    .format(len(self._users)))

    def get_self_id(self):
        """Return UserID of the logged in user."""
        return self._self_user_id

    @gen.coroutine
    def get_self_user(self):
        """Return User who is logged in."""
        yield self.get_users([self._self_user_id])[0]

    @gen.coroutine
    def get_user(self, user_id):
        """Wrapper for get_users for getting a single user."""
        users = yield self.get_users([user_id])
        return users[0]

    @gen.coroutine
    def get_users(self, user_id_list):
        """Gets Users by a list of IDs.

        Returns a dict user_id -> User.

        Until we can find all users reliably, this method will return dummy
        users rather than raising an exception when a user can't be found.
        """
        # Request any new user IDs if necessary
        unknown_user_ids = set(user_id_list) - set(self._users.keys())
        if unknown_user_ids:
            logger.info('Need to request users: {}'.format(unknown_user_ids))
            yield self._request_users(unknown_user_ids)

        # Return Users and add dummies for Users we couldn't find.
        return {user_id: (self._users[user_id] if user_id in self._users
                          else self._make_dummy_user(user_id))
                for user_id in user_id_list}

    def _make_dummy_user(self, user_id):
        """Return a dummy User and add it to the list."""
        logger.info('Creating dummy user for {}'.format(user_id))
        user = parsers.User(id_=user_id, full_name='UNKNOWN',
                            first_name='UNKNOWN',
                            is_self=(user_id == self._self_user_id))
        self._users[user_id] = user
        return user

    @gen.coroutine
    def _request_users(self, user_id_list):
        """Make request for a list of users by ID."""
        res = yield self._client.getentitybyid([user_id.chat_id for user_id
                                                in user_id_list])
        for entity in res['entity']:
            try:
                user = _parse_user_entity(entity)
            except ValueError as e:
                logger.warning('Failed to parse user entity: {}: {}'
                               .format(e, entity))
            else:
                user_id = parsers.UserID(chat_id=user['chat_id'],
                                         gaia_id=user['gaia_id'])
                self._users[user_id] = parsers.User(
                    id_=user_id, full_name=user['full_name'],
                    first_name=user['first_name'],
                    is_self=(user_id == self._self_user_id)
                )


class Client(object):
    """Instant messaging client for Hangouts.

    Maintains a connections to the servers, emits events, and accepts commands.
    """

    def __init__(self, cookies):
        """Create new client.

        cookies is a dictionary of authentication cookies.
        """

        # Event fired when the client connects for the first time with
        # arguments ().
        self.on_connect = event.Event('Client.on_connect')
        # Event fired when the client reconnects after being disconnected with
        # arguments ().
        self.on_reconnect = event.Event('Client.on_reconnect')
        # Event fired when the client is disconnected with arguments ().
        self.on_disconnect = event.Event('Client.on_disconnect')
        # Event fired when a new message arrives with arguments (chat_message).
        self.on_message = event.Event('Client.on_message')
        # Event fired when a user starts or stops typing with arguments
        # (typing_message).
        self.on_typing = event.Event('Client.on_typing')
        # Event fired when a user changes focus on a conversation with
        # arguments (focus_message).
        self.on_focus = event.Event('Client.on_focus')
        # Event fired when a conversation updates with arguments
        # (conversation_message).
        self.on_conversation = event.Event('Client.on_conversation')

        self._cookies = cookies
        self._sync_timestamp = None  # datetime.datetime

        # These are instantiated after ConnectionEvent:
        self.initial_users = None # {UserID: User}
        self.self_user_id = None # UserID
        self.initial_conversations = None # {conv_id: Conversation}

        # hangups.channel.Channel instantiated in connect()
        self._channel = None
        # API key sent with every request:
        self._api_key = None
        # Parameters sent in request headers:
        self._header_date = None
        self._header_version = None
        self._header_id = None
        # TODO This one isn't being set anywhere:
        self._header_client = None
        # Parameters needed to create the Channel:
        self._channel_path = None
        self._clid = None
        self._channel_ec_param = None
        self._channel_prop_param = None

    ##########################################################################
    # Public methods
    ##########################################################################

    @gen.coroutine
    def connect(self):
        """Connect to the server and receive events."""
        yield self._init_talkgadget_1()
        self._channel = channel.Channel(self._cookies, self._channel_path,
                                        self._clid, self._channel_ec_param,
                                        self._channel_prop_param)
        sync_f = lambda: ioloop.IOLoop.instance().add_future(
            self._sync_chat_messages(), lambda f: f.result()
        )
        self._channel.on_connect.add_observer(sync_f)
        self._channel.on_connect.add_observer(self.on_connect.fire)
        self._channel.on_reconnect.add_observer(sync_f)
        self._channel.on_reconnect.add_observer(self.on_reconnect.fire)
        self._channel.on_disconnect.add_observer(self.on_disconnect.fire)
        self._channel.on_message.add_observer(self._on_push_data)
        yield self._channel.listen()

    ##########################################################################
    # Private methods
    ##########################################################################

    @gen.coroutine
    def _sync_chat_messages(self):
        """Sync chat messages since self._sync_timestamp."""
        logger.info('Syncing messages since {}'.format(self._sync_timestamp))
        res = yield self.syncallnewevents(self._sync_timestamp)

        # Parse chat message from response and fire on_message event for each
        # new chat message.
        conversation_state = res[3]
        for conversation in conversation_state:
            events = conversation[2]
            for msg in events:
                try:
                    chat_message = parsers.parse_chat_message([msg])
                except exceptions.ParseError as e:
                    logger.warning('Failed to parse message: {}'.format(e))
                except exceptions.ParseNotImplementedError as e:
                    logger.info('Failed to parse message: {}'.format(e))
                else:
                    # Workaround for syncallnewevents timestamp being
                    # inclusive:
                    if chat_message.timestamp > self._sync_timestamp:
                        self.on_message.fire(chat_message)

        self._sync_timestamp = parsers.from_timestamp(int(res[1][4]))

    @gen.coroutine
    def _init_talkgadget_1(self):
        """Make first talkgadget request and parse response.

        The response body is a HTML document containing a series of script tags
        containing JavaScript object. We need to parse the object to get at the
        data.
        """
        url = 'https://talkgadget.google.com/u/0/talkgadget/_/chat'
        params = {
            'prop': 'aChromeExtension',
            'fid': 'gtn-roster-iframe-id',
            'ec': '["ci:ec",true,true,false]',
        }
        headers = {
            # appears to require a browser user agent
            'user-agent': (
                'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36'
                '(KHTML, like Gecko) Chrome/34.0.1847.132 Safari/537.36'
            ),
        }
        res = yield http_utils.fetch(
            url, cookies=self._cookies, params=params, headers=headers,
            connect_timeout=CONNECT_TIMEOUT, request_timeout=REQUEST_TIMEOUT
        )
        logger.debug('First talkgadget request result:\n{}'.format(res.body))
        if res.code != 200:
            raise ValueError("First talkgadget request failed with {}: {}"
                             .format(res.code, res.body))
        res = res.body.decode()

        # Parse the response by using a regex to find all the JS objects, and
        # parsing them.
        res = res.replace('\n', '')
        regex = re.compile(
            r"(?:<script>AF_initDataCallback\((.*?)\);</script>)"
        )
        data_dict = {}
        for data in regex.findall(res):
            try:
                data = javascript.loads(data)
                # pylint: disable=invalid-sequence-index
                data_dict[data['key']] = data['data']
            except ValueError as e:
                # not everything will be parsable, but we don't care
                logger.debug('Failed to parse JavaScript: {}\n{}'
                             .format(e, data))

        # TODO: handle errors here
        self._api_key = data_dict['ds:7'][0][2]
        self._header_date = data_dict['ds:2'][0][4]
        self._header_version = data_dict['ds:2'][0][6]
        self._header_id = data_dict['ds:4'][0][7]
        self._channel_path = data_dict['ds:4'][0][1]
        self._clid = data_dict['ds:4'][0][7]
        self._channel_ec_param = data_dict['ds:4'][0][4]
        self._channel_prop_param = data_dict['ds:4'][0][5]
        self._sync_timestamp = parsers.from_timestamp(
            data_dict['ds:21'][0][1][4]
        )

        # build dict of conversations and their participants
        initial_conversations = {}
        self.initial_users = {} # {UserID: User}

        # add self to the contacts
        self_contact = data_dict['ds:20'][0][2]
        self.self_user_id = parsers.UserID(chat_id=self_contact[8][0],
                                           gaia_id=self_contact[8][1])
        self.initial_users[self.self_user_id] = parsers.User(
            id_=self.self_user_id, full_name=self_contact[9][1],
            first_name=self_contact[9][2], is_self=True
        )

        conversations = data_dict['ds:19'][0][3]
        for c in conversations:
            id_ = c[1][0][0]
            participants = c[1][13]
            last_modified = c[1][3][12]
            # With every converstion, we get a list of up to 20 of the most
            # recent messages, sorted oldest to newest.
            messages = []
            for raw_message in c[2]:
                try:
                    chat_message = parsers.parse_chat_message([raw_message])
                except exceptions.ParseError as e:
                    logger.warning('Failed to parse message: {}'.format(e))
                except exceptions.ParseNotImplementedError as e:
                    logger.info('Failed to parse message: {}'.format(e))
                else:
                    messages.append(chat_message)
            initial_conversations[id_] = {
                'participants': [],
                'last_modified': last_modified,
                'name': c[1][2],
                'messages': messages,
            }
            # Add the participants for this conversation.
            for p in participants:
                user_id = parsers.UserID(chat_id=p[0][0], gaia_id=p[0][1])
                initial_conversations[id_]['participants'].append(
                    user_id
                )
                # Add the participant to our list of contacts as a fallback, in
                # case they can't be found later by other methods.
                # TODO We should note who these users are and try to request
                # them.
                # p[1] can be a full name, None, or out of range.
                try:
                    display_name = p[1]
                except IndexError:
                    display_name = None
                if display_name is None:
                    display_name = 'Unknown'
                self.initial_users[user_id] = parsers.User(
                    id_=user_id, first_name=display_name.split()[0],
                    full_name=display_name,
                    is_self=(user_id == self.self_user_id)
                )

        # build dict of contacts and their names (doesn't include users not in
        # contacts)
        contacts_main = data_dict['ds:21'][0]
        # contacts_main[2] has some, but the format is slightly different
        contacts = (contacts_main[4][2] + contacts_main[5][2] +
                    contacts_main[6][2] + contacts_main[7][2] +
                    contacts_main[8][2])
        for c in contacts:
            user_id = parsers.UserID(chat_id=c[0][8][0], gaia_id=c[0][8][1])
            self.initial_users[user_id] = parsers.User(
                id_=user_id, full_name=c[0][9][1], first_name=c[0][9][2],
                is_self=(user_id == self.self_user_id)
            )

        # Create a dict of the known conversations.
        self.initial_conversations = {conv_id: Conversation(
            self, conv_id, [self.initial_users[user_id] for user_id
                            in conv_info['participants']],
            conv_info['last_modified'], conv_info['name'],
            conv_info['messages'],
        ) for conv_id, conv_info in initial_conversations.items()}

    def _get_authorization_header(self):
        """Return autorization header for chat API request."""
        # technically, it doesn't matter what the url and time are
        time_msec = int(time.time() * 1000)
        auth_string = '{} {} {}'.format(time_msec, self._get_cookie("SAPISID"),
                                        ORIGIN_URL)
        auth_hash = hashlib.sha1(auth_string.encode()).hexdigest()
        return 'SAPISIDHASH {}_{}'.format(time_msec, auth_hash)

    def _get_cookie(self, name):
        """Return a cookie for raise error if that cookie was not provided."""
        try:
            return self._cookies[name]
        except KeyError:
            raise KeyError("Cookie '{}' is required".format(name))

    def _get_request_header(self):
        """Return request header for chat API request."""
        return [
            [3, 3, self._header_version, self._header_date],
            [self._header_client, self._header_id],
            None,
            "en"
        ]

    def _on_push_data(self, msg_type, msg):
        """Parse channel messages and call the appropriate event."""
        parsed_msg = parsers.parse_message(msg_type, msg)
        # Update the sync timestamp:
        if isinstance(parsed_msg, parsers.ChatMessage):
            self._sync_timestamp = parsed_msg.timestamp
        # Fire the appropriate event:
        handler = {
            parsers.ChatMessage: self.on_message,
            parsers.FocusStatusMessage: self.on_focus,
            parsers.TypingStatusMessage: self.on_typing,
            parsers.ConversationStatusMessage: self.on_conversation,
        }.get(parsed_msg.__class__, None)
        if handler is not None:
            handler.fire(parsed_msg)

    @gen.coroutine
    def _request(self, endpoint, body_json, use_json=True):
        """Make chat API request."""
        url = 'https://clients6.google.com/chat/v1/{}'.format(endpoint)
        headers = {
            'authorization': self._get_authorization_header(),
            'x-origin': ORIGIN_URL,
            'x-goog-authuser': '0',
            'content-type': 'application/json+protobuf',
        }
        required_cookies = ['SAPISID', 'HSID', 'SSID', 'APISID', 'SID']
        cookies = {cookie: self._get_cookie(cookie)
                   for cookie in required_cookies}
        params = {
            'key': self._api_key,
            'alt': 'json' if use_json else 'protojson',
        }
        res = yield http_utils.fetch(
            url, method='POST', headers=headers, cookies=cookies,
            params=params, data=json.dumps(body_json),
            request_timeout=REQUEST_TIMEOUT, connect_timeout=CONNECT_TIMEOUT
        )
        logger.debug('Response to request for {} was {}:\n{}'
                     .format(endpoint, res.code, res.body))
        if res.code != 200:
            raise ValueError('Request to {} endpoint failed with {}: {}'
                             .format(endpoint, res.code, res.body.decode()))
        return res

    @gen.coroutine
    def getselfinfo(self):
        """Return information about your account."""
        res = yield self._request('contacts/getselfinfo', [
            self._get_request_header(),
            [], []
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def setfocus(self, conversation_id):
        """Set focus (occurs whenever you give focus to a client)."""
        res = yield self._request('conversations/setfocus', [
            self._get_request_header(),
            [conversation_id],
            1,
            20
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def searchentities(self, search_string, max_results):
        """Search for people."""
        res = yield self._request('contacts/searchentities', [
            self._get_request_header(),
            [],
            search_string,
            max_results
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def querypresence(self, chat_id):
        """Check someone's presence status."""
        res = yield self._request('presence/querypresence', [
            self._get_request_header(),
            [
                [chat_id]
            ],
            [1, 2, 5, 7, 8]
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def getentitybyid(self, chat_id_list):
        """Return information about a list of contacts."""
        res = yield self._request('contacts/getentitybyid', [
            self._get_request_header(),
            None,
            [[str(chat_id)] for chat_id in chat_id_list],
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def getconversation(self, conversation_id, num_events,
                        storage_continuation_token, event_timestamp):
        """Return data about a conversation.

        Seems to require both a timestamp and a token from a previous event
        """
        res = yield self._request('conversations/getconversation', [
            self._get_request_header(),
            [
                [conversation_id], [], []
            ],
            True, True, None, num_events,
            [None, storage_continuation_token, event_timestamp]
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def syncallnewevents(self, timestamp):
        """List all events occuring at or after timestamp.

        This method requests protojson rather than json so we have one chat
        message parser rather than two.

        timestamp: datetime.datetime instance specifying the time after
        which to return all events occuring in.

        Raises hangups.NetworkError if the request fails.
        """
        try:
            res = yield self._request('conversations/syncallnewevents', [
                self._get_request_header(),
                int(timestamp.timestamp()) * 1000000,
                [], None, [], False, [],
                1048576 # max response size? (number of bytes in a MB)
            ], use_json=False)
        except (httpclient.HTTPError, IOError) as e:
            # In addition to HTTPError, httpclient can raise IOError (which
            # includes socker.gaierror).
            raise exceptions.NetworkError(e)
        # can return 200 but still contain an error
        res = javascript.loads(res.body.decode())
        res_status = res[1][0]
        if res_status != 1:
            raise exceptions.NetworkError('Response status is \'{}\''
                                          .format(res_status))
        return res

    @gen.coroutine
    def syncrecentconversations(self):
        """List the contents of recent conversations, including messages.

        Similar to syncallnewevents, but appears to return a limited number of
        conversations (20) rather than all conversations in a given date range.
        """
        res = yield self._request('conversations/syncrecentconversations', [
            self._get_request_header(),
        ])
        return json.loads(res.body.decode())

    @gen.coroutine
    def setchatname(self, conversation_id, name):
        """Set the name of a conversation.

        Raises hangups.NetworkError if the request can not be sent.
        """
        client_generated_id = random.randint(0, 2 ** 32)
        body = [
            self._get_request_header(),
            None,
            name,
            None,
            [[conversation_id], client_generated_id, 1]
        ]
        try:
            res = yield self._request('conversations/renameconversation', body)
        except (httpclient.HTTPError, IOError) as e:
            # In addition to HTTPError, httpclient can raise IOError (which
            # includes socker.gaierror).
            logger.warning('Failed to send message: {}'.format(e))
            raise exceptions.NetworkError(e)
        res = json.loads(res.body.decode())
        res_status = res['response_header']['status']
        if res_status != 'OK':
            logger.warning('renameconversation returned status {}'
                           .format(res_status))
            raise exceptions.NetworkError()

    @gen.coroutine
    def sendchatmessage(self, conversation_id, message, is_bold=False,
                        is_italic=False, is_strikethrough=False,
                        is_underlined=False):
        """Send a chat message to a conversation.

        message may not be empty.

        Raises hangups.NetworkError if the message can not be sent.
        """
        client_generated_id = random.randint(0, 2**32)
        body = [
            self._get_request_header(),
            None, None, None, [],
            [
                [
                    [0, message, [is_bold, is_italic, is_strikethrough,
                                  is_underlined]]
                ],
                []
            ],
            None,
            [
                [conversation_id], client_generated_id, 2
            ],
            None, None, None, []
        ]
        try:
            res = yield self._request('conversations/sendchatmessage', body)
        except (httpclient.HTTPError, IOError) as e:
            # In addition to HTTPError, httpclient can raise IOError (which
            # includes socker.gaierror).
            logger.warning('Failed to send message: {}'.format(e))
            raise exceptions.NetworkError(e)
        # sendchatmessage can return 200 but still contain an error
        res = json.loads(res.body.decode())
        res_status = res['response_header']['status']
        if res_status != 'OK':
            logger.warning('sendchatmessage returned status {}'
                           .format(res_status))
            raise exceptions.NetworkError()



"""Support for reading messages from the long-polling channel.

Hangouts receives events using a system that appears very close to an App
Engine Channel.
"""

from tornado import gen, httpclient
import logging

from hangups import javascript, parsers, http_utils, event

logger = logging.getLogger(__name__)
# Set the connection and request timeouts low so we fail fast when there's a
# network problem.
CONNECT_TIMEOUT = 10
REQUEST_TIMEOUT = 10
# Long-polling requests may last ~3-4 minutes.
LP_REQ_TIMEOUT = 60 * 5
# Long-polling requests send heartbeats every 15 seconds, so if we miss two in
# a row, consider the connection dead.
LP_DATA_TIMEOUT = 30


def _parse_sid_response(res):
    """Parse response format for request for new channel SID.

    Returns (SID, header_client, gsessionid).
    """
    sid = None
    header_client = None
    gsessionid = None

    p = parsers.PushDataParser()
    res = javascript.loads(list(p.get_submissions(res))[0])
    for segment in res:
        num, message = segment
        if num == 0:
            sid = message[1]
        elif message[0] == 'c':
            type_ = message[1][1][0]
            if type_ == 'cfj':
                header_client = message[1][1][1].split('/')[1]
            elif type_ == 'ei':
                gsessionid = message[1][1][1]

    return(sid, header_client, gsessionid)


class Channel(object):

    """A channel connection that can listen for messages."""

    ##########################################################################
    # Public methods
    ##########################################################################

    def __init__(self, cookies, path, clid, ec, prop):
        """Create a new channel."""

        # Event fired when channel connects with arguments ():
        self.on_connect = event.Event('Channel.on_connect')
        # Event fired when channel reconnects with arguments ():
        self.on_reconnect = event.Event('Channel.on_reconnect')
        # Event fired when channel disconnects with arguments ():
        self.on_disconnect = event.Event('Channel.on_disconnect')
        # Event fired when a channel message is received with arguments
        # (message_type, message):
        self.on_message = event.Event('Channel.on_message')

        # True if the channel is currently connected:
        self._is_connected = False
        # True if the on_connect event has been called at least once:
        self._on_connect_called = False
        # Request cookies dictionary:
        self._cookies = cookies
        # Parser for assembling messages:
        self._push_parser = None

        # Static channel parameters:
        # '/u/0/talkgadget/_/channel/'
        self._channel_path = path
        # 'A672C650270E1674'
        self._clid_param = clid
        # '["ci:ec",1,1,0,"chat_wcs_20140813.110045_RC2"]\n'
        self._ec_param = ec
        # 'aChromeExtension'
        self._prop_param = prop

        # Discovered parameters:
        self._sid_param = None
        self._gsessionid_param = None

    @gen.coroutine
    def listen(self):
        """Listen for messages on the channel.

        This method only returns when the connection has been closed due to an
        error.
        """
        MAX_RETRIES = 5  # maximum number of times to retry after a failure
        retries = MAX_RETRIES # number of remaining retries
        need_new_sid = True  # whether a new SID is needed

        while retries >= 0:
            # After the first failed retry, back off exponentially longer after
            # each attempt.
            if retries + 1 < MAX_RETRIES:
                backoff_seconds = 2 ** (MAX_RETRIES - retries)
                logger.info('Backing off for {} seconds'
                            .format(backoff_seconds))
                yield http_utils.sleep(backoff_seconds)

            # Request a new SID if we don't have one yet, or the previous one
            # became invalid.
            if need_new_sid:
                # TODO: error handling
                yield self._fetch_channel_sid()
                need_new_sid = False
            # Clear any previous push data, since if there was an error it
            # could contain garbage.
            self._push_parser = parsers.PushDataParser()
            try:
                yield self._longpoll_request()
            except IOError as e:
                # An error occurred, so decrement the number of retries.
                retries -= 1
                if self._is_connected:
                    self._is_connected = False
                    self.on_disconnect.fire()
                logger.warning('Long-polling request failed because of '
                               'IOError: {}'.format(e))
            except httpclient.HTTPError as e:
                # An error occurred, so decrement the number of retries.
                retries -= 1
                if self._is_connected:
                    self._is_connected = False
                    self.on_disconnect.fire()
                if e.code == 400 and e.response.reason == 'Unknown SID':
                    logger.warning('Long-polling request failed because SID '
                                   'became invalid. Will attempt to recover.')
                    need_new_sid = True
                elif e.code == 599:
                    logger.warning('Long-polling request failed because '
                                   'connection was closed. Will attempt to '
                                   'recover.')
                else:
                    logger.warning('Long-polling request failed for unknown '
                                   'reason: {}'.format(e))
                    break # Do not retry.
            else:
                # The connection closed successfully, so reset the number of
                # retries.
                retries = MAX_RETRIES

            # If the request ended with an error, the client must account for
            # messages being dropped during this time.

        logger.error('Ran out of retries for long-polling request')

    ##########################################################################
    # Private methods
    ##########################################################################

    @gen.coroutine
    def _fetch_channel_sid(self):
        """Request a new session ID for the push channel."""
        logger.info('Requesting new session...')
        url = 'https://talkgadget.google.com{}bind'.format(self._channel_path)
        params = {
            'VER': 8,
            'clid': self._clid_param,
            'ec': self._ec_param,
            'RID': 81187, # TODO: "request ID"? should probably increment
            # Required if we want our client to be called "AChromeExtension":
            'prop': self._prop_param,
        }
        res = yield http_utils.fetch(
            url, method='POST', cookies=self._cookies, params=params,
            data='count=0', connect_timeout=CONNECT_TIMEOUT,
            request_timeout=REQUEST_TIMEOUT
        )
        logger.debug('Fetch SID response:\n{}'.format(res.body))
        if res.code != 200:
            # TODO use better exception
            raise ValueError("SID fetch request failed with {}: {}"
                             .format(res.code, res.raw.read()))
        # TODO: handle errors here
        self._sid_param, _, self._gsessionid_param = (
            _parse_sid_response(res.body)
        )
        logger.info('New SID: {}'.format(self._sid_param))
        logger.info('New gsessionid: {}'.format(self._gsessionid_param))

    @gen.coroutine
    def _longpoll_request(self):
        """Open a long-polling request to receive push events.

        Raises HTTPError or IOError.
        """
        params = {
            'VER': 8,
            'clid': self._clid_param,
            'prop': self._prop_param,
            'ec': self._ec_param,
            'gsessionid': self._gsessionid_param,
            'RID': 'rpc',
            't': 1, # trial
            'SID': self._sid_param,
            'CI': 0,
        }
        URL = 'https://talkgadget.google.com/u/0/talkgadget/_/channel/bind'
        logger.info('Opening new long-polling request')
        res = yield http_utils.longpoll_fetch(
            URL, params=params, cookies=self._cookies,
            streaming_callback=self._on_push_data,
            connect_timeout=CONNECT_TIMEOUT, request_timeout=LP_REQ_TIMEOUT,
            data_timeout=LP_DATA_TIMEOUT
        )
        return res

    def _on_push_data(self, data_bytes):
        """Parse push data and trigger event methods."""
        logger.debug('Received push data:\n{}'.format(data_bytes))

        # This method is only called when the long-polling request was
        # successful, so use it to trigger connection events if necessary.
        if not self._is_connected:
            if self._on_connect_called:
                self._is_connected = True
                self.on_reconnect.fire()
            else:
                self._on_connect_called = True
                self._is_connected = True
                self.on_connect.fire()

        messages = self._push_parser.get_messages(data_bytes)
        for msg_type, msg in messages:
            logger.debug('Received channel message of type {}: {}'
                         .format(msg_type, msg))
            self.on_message.fire(msg_type, msg)



"""Graphical notifications for the hangups UI.

TODO:
    - Support other notification systems (like terminal bell).
    - Support notify-osd's merged notifications. It appears this would require
    using a dbus library so that each notification comes from the same process.
    - Create notifications for other events like (dis)connection
"""

import html
import logging
import re
import subprocess

logger = logging.getLogger(__name__)
NOTIFY_CMD = [
    'gdbus', 'call', '--session', '--dest', 'org.freedesktop.Notifications',
    '--object-path', '/org/freedesktop/Notifications', '--method',
    'org.freedesktop.Notifications.Notify', 'hangups', '{replaces_id}', '',
    '{sender_name}', '{msg_text}', '[]', '{{}}', ' -1'
]
RESULT_RE = re.compile(r'\(uint32 ([\d]+),\)')


class Notifier(object):

    """Receives events from hangups and creates system notifications.

    This uses the gdbus utility to create freedesktop.org notifications. If a
    new notification is created while a previous one is still open, the
    previous notification is instantly replaced.
    """

    def __init__(self, client, conv_list):
        self._conv_list = conv_list  # hangups.ConversationList
        self._client = client  # hangups.Client
        self._client.on_message.add_observer(self._on_message)
        self._replaces_id = 0

    def _on_message(self, chat_message):
        """Create notification for new messages."""
        conv = self._conv_list.get(chat_message.conv_id)
        user = conv.get_user(chat_message.user_id)
        # Ignore messages sent by yourself.
        if not user.is_self:
            # We have to escape angle brackets because freedesktop.org
            # notifications support markup.
            cmd = [arg.format(
                sender_name=html.escape(user.full_name, quote=False),
                msg_text=html.escape(chat_message.text, quote=False),
                replaces_id=self._replaces_id,
            ) for arg in NOTIFY_CMD]

            # Run the notification and parse out the replaces_id. Since the
            # command is a list of arguments, and we're not using a shell, this
            # should be safe.
            logger.info('Creating notification with command: {}'.format(cmd))
            try:
                output = subprocess.check_output(
                    cmd, stderr=subprocess.STDOUT
                ).decode()
            except (subprocess.CalledProcessError, FileNotFoundError) as e:
                logger.warning('Notification command failed: {}'.format(e))
                return
            try:
                self._replaces_id = RESULT_RE.match(output).groups()[0]
            except (AttributeError, IndexError) as e:
                logger.warning('Failed to parse notification command '
                               'result: {}'.format(e))



"""Exceptions used by hangups."""

class HangupsError(Exception):
    """hangups general exception."""
    pass


class NetworkError(HangupsError):
    """hangups network operation failed."""
    pass


class ParseError(HangupsError):
    """hangups failed to parse a message."""
    pass


class ParseNotImplementedError(HangupsError):
    """hangups parsing is not implemented for this type of message."""
    pass



"""Simple Google auth supporting second factor.

Be careful not to log user credentials here.
"""

import requests
import logging
import getpass
import re
import json


logger = logging.getLogger(__name__)
FORM_RE = re.compile("id=\"(.+?)\"[\s]+value='(.+?)'",re.MULTILINE)
BROWSER_USER_AGENT = (
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 '
    '(KHTML, like Gecko) Chrome/34.0.1847.132 Safari/537.36'
)


class GoogleAuthError(Exception): pass
"""Exception raised when auth fails."""


def _get_galx_token(session):
    """Retrieve GALX token necessary for Google auth."""
    r = session.get((
        "https://accounts.google.com/ServiceLogin?passive=true&skipvpage=true"
        "&continue=https://talkgadget.google.com/talkgadget/"
        "gauth?verify%3Dtrue&authuser=0"
    ))
    if r.status_code != 200:
        raise GoogleAuthError('Failed to get GALX token: request returned {}'
                              .format(r.status_code))
    try:
        galx = r.cookies["GALX"]
    except KeyError:
        raise GoogleAuthError(
            'Failed to get GALX token: cookie was not present'
        )
    return galx


def _send_credentials(session, email, password, galx_token, get_pin_f):
    """Send credentials to Google and return auth cookies."""
    data = {
        "GALX": galx_token,
        "continue": ('https://talkgadget.google.com/talkgadget/'
                     'gauth?verify=true'),
        "skipvpage": 'true',
        "_utf8": '☃',
        "bgresponse": 'js_disabled',
        "pstMsg": '0',
        "dnConn": '',
        "checkConnection": '',
        "checkedDomains": 'youtube',
        "Email": email,
        "Passwd": password,
        "signIn": 'Přihlásit se',
        "PersistentCookie": 'yes',
        "rmShown": '1'
    }
    r = session.post('https://accounts.google.com/ServiceLoginAuth', data=data)
    logger.debug('ServiceLoginAuth response: {} {}'
                 .format(r.status_code, r.url))
    SUCCESS_URL = ('https://talkgadget.google.com/'
                   'talkgadget/gauth?verify=true&pli=1')
    SECONDFACTOR_URL = 'https://accounts.google.com/SecondFactor'
    if r.status_code == 200 and r.url == SUCCESS_URL:
        pass # login success, no second factor
    elif r.status_code == 200 and r.url.startswith(SECONDFACTOR_URL):
        # If credentials are correct, but 2FA is required, it returns redirect
        # to SecondFactor. We have to parse some values out of the form.
        form_values = dict(FORM_RE.findall(r.text))
        try:
            timestamp = form_values['timeStmp']
            sec_token = form_values['secTok']
        except KeyError:
            raise GoogleAuthError(
                'Failed to extract timestamp and sec_token from form.'
            )
        _send_second_factor(session, get_pin_f(), timestamp, sec_token)
    else:
        # something else happened
        raise GoogleAuthError('Login failed for unknown reason')


def _send_second_factor(session, pin, timestamp, sec_token):
    """Send second factor to Google."""
    data = {
        'checkedDomains': 'youtube',
        'checkConnection': 'youtube:73:0',
        'pstMsg': '1',
        'timeStmp': timestamp,
        'secTok': sec_token,
        'smsToken': '',
        'smsUserPin': pin,
        'smsVerifyPin': 'Verify',
        'PersistentOptionSelection': '1',
        'PersistentCookie': 'on',
    }
    r = session.post('https://accounts.google.com/SecondFactor', data=data)
    logger.debug('SecondFactor response: {} {}'
                 .format(r.status_code, r.url))
    SUCCESS_URL = 'https://www.google.com/settings/personalinfo'
    if r.status_code == 200 and r.url == SUCCESS_URL:
        pass # success
    else:
        raise GoogleAuthError('Second factor failed')


def _get_auth_cookies(session):
    """Extract auth cookies from the session."""
    # Just using all the cookies from the session doesn't work in the second
    # factor case.
    req = requests.Request("GET", "https://talkgadget.google.com")
    mock_req = requests.cookies.MockRequest(req)
    cookies = session.cookies._cookies_for_request(mock_req)
    return {c.name: c.value for c in cookies}


def _verify_auth(cookies):
    """Return True if auth cookies are still valid."""
    # TODO: fix this
    # The set of cookies we end up with makes choosing a test URL tricky. I
    # haven't found one what works with and without 2FA.
    return True
    #URL = 'https://clients6.google.com/chat/v1/contacts/getselfinfo'
    #r = requests.post(URL, cookies=cookies, data='')
    #if r.status_code == 401: # unauthorized
    #    logger.info('Auth verification failed: {} {}'
    #                .format(r.status_code, r.url))
    #    return False
    #else:
    #    logger.info('Auth verification succeeded')
    #    return True


def get_auth(get_credentials_f, get_pin_f, cookie_filename):
    """Login into Google and return auth cookies.

    get_credentials_f() is called if credentials are required to log in, and
    should return (email, password). get_pin_f() is called if a pin is required
    to log in, and should return the pin.

    The cookies are saved/loaded from cookie_filename if possible, so
    credentials may not be necessary.

    Raises GoogleAuthError on failure.
    """
    logger.info('Attempting to load auth cookies from {}'
                .format(cookie_filename))
    try:
        with open(cookie_filename) as f:
            cookies = json.load(f)
            if _verify_auth(cookies):
                logger.info('Using saved auth cookies')
                return cookies
            else:
                logger.info('Saved auth cookies failed to verify')
    except (IOError, ValueError):
        logger.info('Failed to load saved auth cookies')

    # Prepare Requests Session with browser-like User-Agent
    session = requests.Session()
    session.headers.update({'User-Agent': BROWSER_USER_AGENT})

    email, password = get_credentials_f()
    galx = _get_galx_token(session)
    _send_credentials(session, email, password, galx, get_pin_f)
    cookies = _get_auth_cookies(session)

    logger.info('Attempting to save auth cookies to {}'
                .format(cookie_filename))
    try:
        with open(cookie_filename, 'w') as f:
            json.dump(cookies, f)
        logger.info('Auth cookies saved')
    except IOError as e:
        logger.warning('Failed to save auth cookies: {}'.format(e))
    return cookies


def get_auth_stdin(cookie_filename):
    """Wrapper for get_auth that prompts the user on stdin."""
    def get_credentials_f():
        """Prompt for and return credentials."""
        email = input('Email: ')
        password = getpass.getpass()
        return (email, password)
    def get_pin_f():
        """Prompt for and return PIN."""
        return input('PIN: ')
    return get_auth(get_credentials_f, get_pin_f, cookie_filename)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    print(get_auth_stdin('cookies.json'))



"""Utility functions for HTTP requests."""

from tornado import gen, concurrent, ioloop, httputil, httpclient
import datetime
import http
import logging

logger = logging.getLogger(__name__)


@concurrent.return_future
def sleep(seconds, callback=None):
    """Return a future that finishes after a given number of seconds."""
    ioloop.IOLoop.instance().add_timeout(datetime.timedelta(seconds=seconds),
                                         callback)


class TimeoutFuture(concurrent.Future):

    """Future that wraps another future so it can be run with a timeout.

    Useful for request timeouts and backoffs.
    """

    def __init__(self, future, timeout_seconds):
        super().__init__()
        future.add_done_callback(self.done_callback)
        sleep(timeout_seconds).add_done_callback(self.done_callback)

    def done_callback(self, future):
        self.set_result(future)


@gen.coroutine
def fetch(url, method='GET', params=None, headers=None, cookies=None,
          data=None, streaming_callback=None, header_callback=None,
          connect_timeout=None, request_timeout=None):
    """Make an HTTP request and return a Future.

    This is mostly just a wrapper for tornado.httpclient, but adds support for
    sending cookies.

    Raises HTTPError and IOError.
    """
    if headers is None:
        headers = {}
    if params is not None:
        url = httputil.url_concat(url, params)
    if cookies is not None:
        # abuse SimpleCookie to escape our cookies for us
        simple_cookies = http.cookies.SimpleCookie(cookies)
        headers['cookie'] = '; '.join(val.output(header='')[1:]
                                      for val in simple_cookies.values())
    http_client = httpclient.AsyncHTTPClient()
    res = yield http_client.fetch(httpclient.HTTPRequest(
        httputil.url_concat(url, params), method=method,
        headers=httputil.HTTPHeaders(headers), body=data,
        streaming_callback=streaming_callback, header_callback=header_callback,
        connect_timeout=connect_timeout, request_timeout=request_timeout
    ))
    return res


@gen.coroutine
def longpoll_fetch(url, method='GET', params=None, headers=None, cookies=None,
                   data=None, streaming_callback=None, connect_timeout=None,
                   request_timeout=None, data_timeout=None):
    """Make a long-polling request and return a Future.

    The future finishes when the request is closed. If the response code is
    2xx, streaming_callback is called with byte data as it arrives.

    The data_timeout is the number of seconds which may elapse with no data
    being received before the connection is considered dead.

    Raises HTTPError and IOError.
    """
    error = False
    new_data = False
    response_code = None  # Persistent state for the header callback

    def _streaming_callback(data):
        """Internal callback for response data."""
        nonlocal error, new_data
        # Call the real streaming_callback only if no error has occurred.
        if error:
            logger.info('Long-polling request ignoring data because of error')
        else:
            logger.info('Long-polling request received data')
            new_data = True
            streaming_callback(data)

    def _header_callback(line):
        """Internal callback for header lines."""
        nonlocal error, response_code
        # If this is the first line of the response, parse it for the response
        # code. An error has occured if the response_code is not 2xx.
        if response_code == None:
            response_code = int(line.split(' ')[1])
            logger.info('Long-polling request received response start, code={}'
                        .format(response_code))
            if response_code >= 300 or response_code < 200:
                error = True

    logger.info('Starting long-polling request: {}'.format(url))
    future = fetch(
        url, method=method, params=params, headers=headers, cookies=cookies,
        data=data, streaming_callback=_streaming_callback,
        header_callback=_header_callback, connect_timeout=connect_timeout,
        request_timeout=request_timeout
    )
    while not future.done():
        # Wait until the request finishes, or the timeout expires, whichever
        # happens first.
        if data_timeout:
            yield TimeoutFuture(future, data_timeout)
        else:
            yield future
        # If the timeout expired and no new data arrived, the request has timed
        # out.
        if new_data or future.done():
            new_data = False
        else:
            logger.info('Long-polling request timed out')
            error = True
            raise IOError('timed out waiting for data')
    logger.info('Long-polling request finished')
    return future.result()



"""Parser for long-polling responses from the talkgadget API."""

import logging
import re
from collections import namedtuple
import datetime

from hangups import javascript, exceptions


logger = logging.getLogger(__name__)
LEN_REGEX = re.compile(r'([0-9]+)\n', re.MULTILINE)


User = namedtuple('User', ['id_', 'full_name', 'first_name', 'is_self'])


UserID = namedtuple('UserID', ['chat_id', 'gaia_id'])


def _best_effort_decode(data_bytes):
    """Decode data_bytes into a string using UTF-8.

    If data_bytes cannot be decoded, pop the last byte until it can be or
    return an empty string.
    """
    for end in reversed(range(1, len(data_bytes) + 1)):
        try:
            return data_bytes[0:end].decode()
        except UnicodeDecodeError:
            pass
    return ''


class PushDataParser(object):
    """Parse data from the long-polling endpoint."""

    def __init__(self):
        # Buffer for bytes containing utf-8 text:
        self._buf = b''

    def get_submissions(self, new_data_bytes):
        """Yield submissions generated from received data.

        Responses from the push endpoint consist of a sequence of submissions.
        Each submission is prefixed with its length followed by a newline.

        The buffer may not be decodable as UTF-8 if there's a split multi-byte
        character at the end. To handle this, do a "best effort" decode of the
        buffer to decode as much of it as possible.

        The length is actually the length of the string as reported by
        JavaScript. JavaScript's string length function returns the number of
        code units in the string, represented in UTF-16. We can emulate this by
        encoding everything in UTF-16 and multipling the reported length by 2.

        Note that when encoding a string in UTF-16, Python will prepend a
        byte-order character, so we need to remove the first two bytes.
        """
        self._buf += new_data_bytes

        while True:

            buf_decoded = _best_effort_decode(self._buf)
            buf_utf16 = buf_decoded.encode('utf-16')[2:]

            lengths = LEN_REGEX.findall(buf_decoded)
            if len(lengths) == 0:
                break
            else:
                # Both lengths are in number of bytes in UTF-16 encoding.
                # The length of the submission:
                length = int(lengths[0]) * 2
                # The length of the submission length and newline:
                length_length = len((lengths[0] + '\n').encode('utf-16')[2:])
                if len(buf_utf16) - length_length < length:
                    break

                submission = buf_utf16[length_length:length_length + length]
                yield submission.decode('utf-16')
                # Drop the length and the submission itself from the beginning
                # of the buffer.
                drop_length = (len((lengths[0] + '\n').encode()) +
                               len(submission.decode('utf-16').encode()))
                self._buf = self._buf[drop_length:]

    def get_messages(self, new_data_bytes):
        """Yield (message_type, message) tuples from received data."""
        # One submission may contain multiple messages.
        for submission in self.get_submissions(new_data_bytes):
            # For each submission payload, yield its messages
            for payload in _get_submission_payloads(submission):
                if payload is not None:
                    yield from _parse_payload(payload)


def _get_submission_payloads(submission):
    """Yield a submission's payloads.

    Most submissions only contain one payload, but if the long-polling
    connection was closed while something happened, there can be multiple
    payloads.
    """
    for sub in javascript.loads(submission):

        # the submission number, increments with each payload
        # sub_num = sub[0]
        # the submission type
        sub_type = sub[1][0]

        if sub_type == 'c':

            # session ID, should be the same for every request
            # session_id = sub[1][1][0]
            # payload type
            payload_type = sub[1][1][1][0]

            if payload_type == 'bfo':
                # Payload is submessages in the list format. These are the
                # payloads we care about.
                yield javascript.loads(sub[1][1][1][1])
            elif payload_type == 'tm':
                # Payload is object format. I'm not sure what these are for,
                # but they don't seem very important.
                pass
            elif payload_type == 'wh':
                # Payload is null. These messages don't contain any information
                # other than the session_id, and appear to be just heartbeats.
                pass
            elif payload_type == 'otr':
                # Not sure what this is for, might be something to do with
                # XMPP.
                pass
            elif payload_type == 'ho:hin':
                # Sent when a video call starts/stops.
                pass
            else:
                logger.warning(
                    'Got submission with unknown payload type {}:\n{}'
                    .format(payload_type, sub)
                )
        elif sub_type == 'noop':
            # These contain no information and only seem to appear once as the
            # first message when a channel is opened.
            pass
        else:
            logger.warning('Got submission with unknown submission type: {}\n{}'
                           .format(sub_type, sub))


def _parse_payload(payload):
    """Parse the list payload format into messages."""
    # the payload begins with a constant header
    if payload[0] != 'cbu':
        raise ValueError('Invalid list payload header: {}'.format(payload[0]))

    # The first submessage is always present, so let's treat it like a header.
    # It doesn't have anything we need so ignore it.
    # first_submsg = payload[1][0][0]

    # The type of a submessage is determined by its position in the array
    yield from ((msg_type, msg) for msg_type, msg in
                enumerate(payload[1][0][1:]) if msg is not None)


def parse_message(message_type, message):
    """Parse any type of message.

    Returns message instance, or None if the message could not be parsed.
    """
    def parse_not_implemented_message(message_type, message):
        """Parser for known but not implemented messages."""
        raise exceptions.ParseNotImplementedError(
            'Unimplemented message type: {}'.format(message_type)
        )
    def parse_unknown_message(message_type, message):
        """Parser for unknown messages."""
        raise exceptions.ParseError(
            'Unknown message type: {}'.format(message_type)
        )

    # Message types have been observed to range from 1 to 14.
    # TODO: Implement the unimplemented message parsers.
    MESSAGE_PARSERS = {
        1: parse_chat_message,
        2: parse_focus_status_message,
        3: parse_typing_status_message,
        # read state change
        6: lambda msg: parse_not_implemented_message(6, msg),
        11: parse_conversation_status_message,
        # notification snooze / set mood
        12: lambda msg: parse_not_implemented_message(12, msg),
        # ??? (a user ID and some properties)
        14: lambda msg: parse_not_implemented_message(14, msg),
    }

    parsed_message = None

    try:
        parser = MESSAGE_PARSERS.get(
            message_type, lambda msg: parse_unknown_message(message_type, msg)
        )
        parsed_message = parser(message)
    except exceptions.ParseError as e:
        logger.warning('Failed to parse message: {}'.format(e))
    except exceptions.ParseNotImplementedError as e:
        logger.info('Failed to parse message: {}'.format(e))

    logger.debug('Parsed message: {}'.format(parsed_message))
    return parsed_message


##############################################################################
# Message parsing utils
##############################################################################


def from_timestamp(timestamp):
    """Convert a microsecond timestamp to a UTC datetime instance."""
    return datetime.datetime.fromtimestamp(timestamp / 1000000,
                                           datetime.timezone.utc)


##############################################################################
# Message types and parsers
##############################################################################


ChatMessage = namedtuple(
    'ChatMessage', ['conv_id', 'user_id', 'timestamp', 'text']
)


def parse_chat_message(message):
    """Return ChatMessage from parsing raw message.

    Raises ParseError if the message cannot be parsed.
    """
    # The message content is a list of message segments, which have a
    # type. For now, let's ignore the types and just use the textual
    # representation, appending all the segments into one string.
    message_text = ''

    # Find the message type:
    type_int = message[0][22]
    try:
        type_ = {
            # message[0][6] will contain message content.
            1: 'REGULAR_CHAT_MESSAGE',
            # message[0][8] will contain ???.
            4: 'UNKNOWN_4',
            # message[0][8] will contain ???.
            5: 'UNKNOWN_5',
            # message[0][9] will contain conversation rename.
            6: 'RENAME_CONVERSATION',
            # message[0][10] will contain hangout event.
            7: 'HANGOUT_EVENT',
            # message[0][13] will contain OTR modification.
            9: 'OTR_MODIFICATION',
        }[type_int]
    except KeyError:
        logger.debug(message)
        raise exceptions.ParseError('Unknown chat message type: {}'
                                    .format(type_int))

    if type_ != 'REGULAR_CHAT_MESSAGE':
        raise exceptions.ParseNotImplementedError(
            'Unimplemented chat message type: {}'.format(type_)
        )

    message_content = message[0][6][2][0]
    if len(message_content) > 0:
        for segment in message_content:
            # known types: 0: text, 1: linebreak, 2: link
            # Hangouts for Android (unlike the Chrome extension) doesn't set
            # the message text for linebreaks, so we have to handle that case
            # separately.
            if segment[0] == 1:
                message_text += '\n'
            else:
                message_text += segment[1]
    else:
        # Try to parse an image message. Image messages contain no message
        # segments, and thus have no automatic textual fallback.
        try:
            # set the message text to the image URL
            message_text = list(message[0][6][2][1][0][0][1].values())[0][0][3]
        except (TypeError, IndexError, ValueError) as e:
            raise exceptions.ParseError('Failed to parse image message: {}'
                                        .format(e))
    return ChatMessage(
        conv_id=message[0][0][0],
        user_id=UserID(chat_id=message[0][1][0], gaia_id=message[0][1][0]),
        timestamp=from_timestamp(message[0][2]),
        text=message_text,
    )


ConversationStatusMessage = namedtuple(
    'ConversationStatusMessage', ['conv_id', 'user_id_list']
)


def parse_conversation_status_message(message):
    """Return ConversationStatusMessage from parsing raw message.

    Raises ParseError if the message cannot be parsed.
    """
    # TODO: This is far from a complete parse.
    return ConversationStatusMessage(
        conv_id=message[0][0],
        user_id_list=[UserID(chat_id=item[0][0], gaia_id=item[0][1])
                      for item in message[13]],
    )


FocusStatusMessage = namedtuple(
    'FocusStatusMessage',
    ['conv_id', 'user_id', 'timestamp', 'status', 'device']
)


def parse_focus_status_message(message):
    """Return FocusStatusMessage from parsing raw message.

    Raises ParseError if the message cannot be parsed.
    """
    FOCUS_STATUSES = {
        1: 'focused',
        2: 'unfocused',
    }
    try:
        focus_status = FOCUS_STATUSES[message[3]]
    except KeyError:
        raise exceptions.ParseError('Unknown focus status: {}'
                                    .format(message[3]))

    FOCUS_DEVICES = {
        20: 'desktop',
        300: 'mobile',
        None: 'unspecified',
    }
    try:
        # sometimes the device is unspecified so the message is shorter
        focus_device = FOCUS_DEVICES[message[4] if len(message) > 4 else None]
    except KeyError:
        raise exceptions.ParseError('Unknown focus device: {}'
                                    .format(message[4]))

    return FocusStatusMessage(
        conv_id=message[0][0],
        user_id=UserID(chat_id=message[1][0], gaia_id=message[1][1]),
        timestamp=from_timestamp(message[2]),
        status=focus_status,
        device=focus_device,
    )


TypingStatusMessage = namedtuple(
    'TypingStatusMessage', ['conv_id', 'user_id', 'timestamp', 'status']
)


def parse_typing_status_message(message):
    """Return TypingStatusMessage from parsing raw message.

    The same status may be sent multiple times consecutively, and when a
    message is sent the typing status will not change to stopped.

    Raises ParseError if the message cannot be parsed.
    """
    TYPING_STATUSES = {
        1: 'typing', # the user is typing
        2: 'paused', # the user stopped typing with inputted text
        3: 'stopped', # the user stopped typing with no inputted text
    }
    try:
        typing_status = TYPING_STATUSES[message[3]]
    except KeyError:
        raise exceptions.ParseError('Unknown typing status: {}'
                                    .format(message[3]))
    return TypingStatusMessage(
        conv_id=message[0][0],
        user_id=UserID(chat_id=message[1][0], gaia_id=message[1][1]),
        timestamp=from_timestamp(message[2]),
        status=typing_status,
    )



"""Demo chat client using Hangups."""

from tornado import ioloop
import appdirs
import argparse
import datetime
import logging
import os
import sys
import urwid

import hangups
from hangups.notify import Notifier


LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
COL_SCHEMES = {
    # Very basic scheme with no colour
    'default': {
        ('active_tab', '', ''),
        ('inactive_tab', 'standout', ''),
        ('msg_date', '', ''),
        ('msg_sender', '', ''),
        ('msg_text', '', ''),
        ('status_line', 'standout', ''),
        ('tab_background', 'standout', ''),
    },
    'solarized-dark': {
        ('active_tab', 'light gray', 'light blue'),
        ('inactive_tab', 'underline', 'light green'),
        ('msg_date', 'dark cyan', ''),
        ('msg_sender', 'dark blue', ''),
        ('msg_text', '', ''),
        ('status_line', 'standout', ''),
        ('tab_background', 'underline', 'black'),
    },
}


class ChatUI(object):
    """User interface for hangups."""

    def __init__(self, cookies_path, keybindings, palette):
        """Start the user interface."""
        self._keys = keybindings

        # These are populated by on_connect when it's called.
        self._conv_widgets = {} # {conversation_id: ConversationWidget}
        self._tabbed_window = None # TabbedWindowWidget
        self._conv_list = None # hangups.ConversationList
        self._user_list = None # hangups.UserList
        self._notifier = None # hangups.notify.Notifier

        # TODO Add urwid widget for getting auth.
        try:
            cookies = hangups.auth.get_auth_stdin(cookies_path)
        except hangups.GoogleAuthError as e:
            print('Login failed ({})'.format(e))
            sys.exit(1)

        self._client = hangups.Client(cookies)
        self._client.on_connect.add_observer(self._on_connect)
        self._client.on_disconnect.add_observer(self._on_disconnect)
        self._client.on_message.add_observer(self._on_message)

        class MyEventLoop(urwid.TornadoEventLoop):
            """Patched Tornado event loop for urwid.

            Patch urwid's TornadoEventLoop to run_sync() our connection
            coroutine rather than calling start(). This "fixes" exception
            handling, causing all uncaught exceptions to be fatal.
            """
            _client = self._client
            def run(self):
                try:
                    ioloop.IOLoop.instance().run_sync(self._client.connect)
                except ioloop.TimeoutError:
                    # Ignore spurious timeout when there's another exception.
                    pass
                # Raise exception if there was one.
                if self._exception:
                    exc, self._exception = self._exception, None
                    raise exc # pylint: disable=E0702

        # Initialize urwid, starting the IOLoop, and block until the IOLoop
        # exits.
        self._urwid_loop = urwid.MainLoop(LoadingWidget(), palette,
                                          event_loop=MyEventLoop(),
                                          handle_mouse=False)
        self._urwid_loop.run()

    def get_conv_widget(self, conv_id):
        """Return an existing or new ConversationWidget."""
        if conv_id not in self._conv_widgets:
            set_title_cb = (lambda widget, title:
                            self._tabbed_window.set_tab(widget, title=title))
            widget = ConversationWidget(self._client,
                                        self._conv_list.get(conv_id),
                                        set_title_cb)
            self._conv_widgets[conv_id] = widget
        return self._conv_widgets[conv_id]

    def add_conversation_tab(self, conv_id, switch=False):
        """Add conversation tab if not present, and optionally switch to it."""
        conv_widget = self.get_conv_widget(conv_id)
        self._tabbed_window.set_tab(conv_widget, switch=switch)

    def on_select_conversation(self, conv_id):
        """Called when the user selects a new conversation to listen to."""
        # switch to new or existing tab for the conversation
        self.add_conversation_tab(conv_id, switch=True)

    def _on_connect(self):
        """Handle connecting for the first time."""
        self._conv_list = hangups.ConversationList(self._client)
        self._user_list = hangups.UserList(self._client)
        self._notifier = Notifier(self._client, self._conv_list)
        # show the conversation menu
        conv_picker = ConversationPickerWidget(self._conv_list,
                                               self.on_select_conversation)
        self._tabbed_window = TabbedWindowWidget(self._keys)
        self._tabbed_window.set_tab(conv_picker, switch=True,
                                    title='Conversations')
        self._urwid_loop.widget = self._tabbed_window

    def _on_message(self, chat_message):
        """Open conversation tab for new messages when they arrive."""
        self.add_conversation_tab(chat_message.conv_id)

    def _on_disconnect(self):
        """Handle disconnecting."""
        # TODO: handle this
        print('Connection lost')


def get_conv_name(conv, truncate=False):
    """Return the readable name for a conversation.

    For one-to-one conversations, the name is the full name of the other user.
    For group conversations, the name is a comma-separated list of first names.

    If truncate is true, only show up to two names in a group conversation.
    """
    participants = sorted(user for user in conv.users if not user.is_self)
    names = [user.first_name for user in participants]
    if len(participants) == 1:
        return participants[0].full_name
    elif truncate and len(participants) > 2:
        return ', '.join(names[:2] + ['+{}'.format(len(names) - 2)])
    else:
        return ', '.join(names)


class LoadingWidget(urwid.WidgetWrap):
    """Widget that shows a loading indicator."""

    def __init__(self):
        # show message in the center of the screen
        super().__init__(urwid.Filler(
            urwid.Text('Connecting...', align='center')
        ))


class ConversationPickerWidget(urwid.WidgetWrap):
    """Widget for picking a conversation."""

    def __init__(self, conversation_list, on_select):
        # Build buttons for selecting conversations ordered by most recently
        # modified first.
        convs = sorted(conversation_list.get_all(), reverse=True,
                       key=lambda c: c.last_modified)
        on_press = lambda button, conv_id: on_select(conv_id)
        buttons = [urwid.Button(get_conv_name(conv), on_press=on_press,
                                user_data=conv.id_)
                   for conv in convs]
        listbox = urwid.ListBox(urwid.SimpleFocusListWalker(buttons))
        widget = urwid.Padding(listbox, left=2, right=2)
        super().__init__(widget)


class ReturnableEdit(urwid.Edit):
    """Edit widget that clears itself and calls a function on return."""

    def __init__(self, on_return, caption=None):
        super().__init__(caption=caption)
        self._on_return = on_return

    def keypress(self, size, key):
        key = super().keypress(size, key)
        if key == 'enter':
            self._on_return(self.get_edit_text())
            self.set_edit_text('')
        else:
            return key


class StatusLineWidget(urwid.WidgetWrap):
    """Widget for showing typing status."""

    def __init__(self, conversation):
        self._typing_statuses = {}
        self._conversation = conversation
        self._conversation.on_message.add_observer(self._on_message)
        self._conversation.on_typing.add_observer(self._on_typing)
        self._widget = urwid.Text('', align='center')
        super().__init__(urwid.AttrWrap(self._widget, 'status_line'))

    def _on_message(self, chat_message):
        """Make users stop typing when they send a message."""
        self._typing_statuses[chat_message.user_id] = 'stopped'
        self._update()

    def _on_typing(self, typing_message):
        """Handle typing updates."""
        self._typing_statuses[typing_message.user_id] = typing_message.status
        self._update()

    def _update(self):
        """Update list of typers."""
        typers = [self._conversation.get_user(user_id).first_name
                  for user_id, status in self._typing_statuses.items()
                  if status == 'typing']
        if len(typers) > 0:
            msg = '{} {} typing...'.format(
                ', '.join(sorted(typers)),
                'is' if len(typers) == 1 else 'are'
            )
        else:
            msg = ''
        self._widget.set_text(msg)


class MessageWidget(urwid.WidgetWrap):

    """Widget for displaying a single message in a conversation."""

    def __init__(self, timestamp, text, user=None):
        # Save the timestamp as an attribute for sorting.
        self.timestamp = timestamp
        text = [
            ('msg_date', '(' + self._get_date_str(timestamp) + ') '),
            ('msg_text', text)
        ]
        if user is not None:
            text.insert(1, ('msg_sender', user.first_name + ': '))
        self._widget = urwid.Text(text)
        super().__init__(self._widget)

    @staticmethod
    def _get_date_str(timestamp):
        """Convert UTC datetime into user interface string."""
        return timestamp.astimezone(tz=None).strftime('%I:%M:%S %p')

    def __lt__(self, other):
        return self.timestamp < other.timestamp


class ConversationWidget(urwid.WidgetWrap):
    """Widget for interacting with a conversation."""

    def __init__(self, client, conversation, set_title_cb):
        client.on_disconnect.add_observer(lambda: self._show_info_message(
            'Disconnected.'
        ))
        client.on_reconnect.add_observer(lambda: self._show_info_message(
            'Connected.'
        ))
        self._conversation = conversation
        self._conversation.on_message.add_observer(self._on_message)

        self._num_unread = 0
        self._set_title_cb = set_title_cb
        self._set_title()

        self._list_walker = urwid.SimpleFocusListWalker([])
        self._list_box = urwid.ListBox(self._list_walker)
        self._status_widget = StatusLineWidget(conversation)
        self._widget = urwid.Pile([
            ('weight', 1, self._list_box),
            ('pack', self._status_widget),
            ('pack', ReturnableEdit(self._on_return, caption='Send message: ')),
        ])
        # focus the edit widget by default
        self._widget.focus_position = 2

        # Display any old messages already attached to the conversation.
        for chat_message in self._conversation.chat_messages:
            self._on_message(chat_message)
        self._num_unread = 0
        self._set_title()

        super().__init__(self._widget)

    def keypress(self, size, key):
        """Handle keypresses marking messages as read."""
        self._num_unread = 0
        self._set_title()
        return super().keypress(size, key)

    def _set_title(self):
        """Update this conversation's tab title."""
        title = get_conv_name(self._conversation, truncate=True)
        if self._num_unread > 0:
            title += ' ({})'.format(self._num_unread)
        self._set_title_cb(self, title)

    def _on_return(self, text):
        """Called when the user presses return on the send message widget."""
        # Ignore if the user hasn't typed a message.
        if len(text) == 0:
            return
        # XXX: Exception handling here is still a bit broken. Uncaught
        # exceptions in _on_message_sent will only be logged.
        self._conversation.send_message(text).add_done_callback(
            self._on_message_sent
        )

    def _on_message_sent(self, future):
        """Handle showing an error if a message fails to send."""
        try:
            future.result()
        except hangups.NetworkError:
            self._show_info_message('Failed to send message.')

    def _add_message_widget(self, message_widget):
        """Add MessageWidget to the ConversationWidget.

        Automatically scroll down to show the new text if the bottom is showing
        (the last widget is focused). This allows the user to scroll up to read
        previous messages while new messages are arriving.
        """
        try:
            bottom_visible = (self._list_box.focus_position ==
                              len(self._list_walker) - 1)
        except IndexError:
            bottom_visible = True  # ListBox is empty
        self._list_walker.append(message_widget)
        self._list_walker.sort()
        if bottom_visible:
            # set_focus_valign is necessary so the last message is always shown
            # completely.
            self._list_box.set_focus(len(self._list_walker) - 1)
            self._list_box.set_focus_valign('bottom')

    def _show_info_message(self, text):
        """Display an informational message with timestamp."""
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self._add_message_widget(MessageWidget(timestamp, text, None))

    def _on_message(self, chat_message):
        """Display a new conversation message."""
        user = self._conversation.get_user(chat_message.user_id)

        # Format the message and add it to the list box.
        self._add_message_widget(MessageWidget(chat_message.timestamp,
                                               chat_message.text, user))

        # Update the count of unread messages.
        if not user.is_self:
            self._num_unread += 1
            self._set_title()


class TabbedWindowWidget(urwid.WidgetWrap):

    """A widget that displays a list of widgets via a tab bar."""

    def __init__(self, keybindings):
        self._widgets = [] # [urwid.Widget]
        self._widget_title = {} # {urwid.Widget: str}
        self._tab_index = None # int
        self._keys = keybindings
        self._tabs = urwid.Text('')
        self._frame = urwid.Frame(None)
        super().__init__(urwid.Pile([
            ('pack', urwid.AttrWrap(self._tabs, 'tab_background')),
            ('weight', 1, self._frame),
        ]))

    def _update_tabs(self):
        """Update tab display."""
        text = []
        for num, widget in enumerate(self._widgets):
            palette = ('active_tab' if num == self._tab_index
                       else 'inactive_tab')
            text += [
                (palette, ' {} '.format(self._widget_title[widget]).encode()),
                ('tab_background', b' '),
            ]
        self._tabs.set_text(text)
        self._frame.contents['body'] = (self._widgets[self._tab_index], None)

    def keypress(self, size, key):
        """Handle keypresses for changing tabs."""
        key = super().keypress(size, key)
        num_tabs = len(self._widgets)
        if key == self._keys['prev_tab']:
            self._tab_index = (self._tab_index - 1) % num_tabs
            self._update_tabs()
        elif key == self._keys['next_tab']:
            self._tab_index = (self._tab_index + 1) % num_tabs
            self._update_tabs()
        else:
            return key

    def set_tab(self, widget, switch=False, title=None):
        """Add or modify a tab.

        If widget is not a tab, it will be added. If switch is True, switch to
        this tab. If title is given, set the tab's title.
        """
        if widget not in self._widgets:
            self._widgets.append(widget)
            self._widget_title[widget] = ''
        if switch:
            self._tab_index = self._widgets.index(widget)
        if title:
            self._widget_title[widget] = title
        self._update_tabs()


def main():
    """Main entry point."""
    # Build default paths for files.
    dirs = appdirs.AppDirs('hangups', 'hangups')
    default_log_path = os.path.join(dirs.user_log_dir, 'hangups.log')
    default_cookies_path = os.path.join(dirs.user_cache_dir, 'cookies.json')

    parser = argparse.ArgumentParser(
        prog='hangups', formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    dirs = appdirs.AppDirs('hangups', 'hangups')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='log detailed debugging messages')
    parser.add_argument('--log', default=default_log_path,
                        help='log file path')
    parser.add_argument('--cookies', default=default_cookies_path,
                        help='cookie storage path')
    parser.add_argument('--key-next-tab', default='ctrl d',
                        help='keybinding for next tab')
    parser.add_argument('--key-prev-tab', default='ctrl u',
                        help='keybinding for previous tab')
    parser.add_argument('--col-scheme', choices=COL_SCHEMES.keys(),
                        default='default', help='colour scheme to use')
    args = parser.parse_args()

    # Create all necessary directories.
    for path in [args.log, args.cookies]:
        directory = os.path.dirname(path)
        if directory != '' and not os.path.isdir(directory):
            try:
                os.makedirs(directory)
            except OSError as e:
                sys.exit('Failed to create directory: {}'.format(e))

    log_level = logging.DEBUG if args.debug else logging.WARNING
    logging.basicConfig(filename=args.log, level=log_level, format=LOG_FORMAT)

    try:
        ChatUI(args.cookies, {
            'next_tab': args.key_next_tab,
            'prev_tab': args.key_prev_tab,
        }, COL_SCHEMES[args.col_scheme])
    except KeyboardInterrupt:
        pass
    except:
        # urwid will prevent some exceptions from being printed unless we use
        # print a newline first.
        print('')
        raise


if __name__ == '__main__':
    main()



"""Parser for a subset of JavaScript written with purplex.

Parses a broader subset of JavaScript than just JSON, needed for parsing some
API responses. This is only as complete as necessary to parse the responses
we're getting.
"""

import purplex


def loads(string):
    """Parse simple JavaScript types from string into Python types.

    Raises ValueError if parsing fails.
    """
    try:
        return _PARSER.parse(string)
    except purplex.exception.PurplexError as e:
        raise ValueError('Failed to load JavaScript: {}'.format(e))


# TODO: there are more possible escape sequences
_ESCAPES = {
    "'": "'",
    '"': '"',
    '\\': '\\',
    'n': '\n',
    't': '\t',
    'r': '\r',
    'u': '', # unicode escapes are a special case
}
_STRING_RE = ('(\'(([^\\\\\'])|(\\\\[{0}]))*?\')|("(([^\\\\"])|(\\\\[{0}]))*?")'
              .format(''.join(_ESCAPES.keys()).replace('\\', '\\\\')))


def _unescape_string(s):
    """Unescape JavaScript escape sequences."""
    chars = list(s)
    unescaped_chars = []
    while len(chars) > 0:
        c = chars.pop(0)
        if c != '\\':
            unescaped_chars.append(c)
        else:
            try:
                c = chars.pop(0)
            except IndexError:
                raise ValueError('Reached end of string literal '
                                 'prematurely: {}'.format(s))
            if c == 'u':
                try:
                    unescaped_chars.append(
                        chr(int(''.join([chars.pop(0) for _ in range(4)]), 16))
                    )
                except IndexError:
                    raise ValueError('Reached end of string literal '
                                     'prematurely: {}'.format(s))
            else:
                try:
                    unescaped_chars.append(_ESCAPES[c])
                except KeyError:
                    raise ValueError('String literal contains invalid '
                                     'escape sequence: {}'.format(s))
    return "".join(unescaped_chars)


class JavaScriptLexer(purplex.Lexer):
    """Lexer for a subset of JavaScript."""
    # TODO negatives? floats?
    INTEGER = purplex.TokenDef(r'\d+')
    FLOAT = purplex.TokenDef(r'[-+]?\d*[.]\d+')

    NULL = purplex.TokenDef(r'null')
    TRUE = purplex.TokenDef(r'true')
    FALSE = purplex.TokenDef(r'false')

    LIST_START = purplex.TokenDef(r'\[')
    LIST_END = purplex.TokenDef(r'\]')
    OBJECT_START = purplex.TokenDef(r'\{')
    OBJECT_END = purplex.TokenDef(r'\}')
    COMMA = purplex.TokenDef(r',')
    COLON = purplex.TokenDef(r':')

    STRING = purplex.TokenDef(_STRING_RE)
    # TODO more unquoted keys are allowed
    KEY = purplex.TokenDef(r'[a-zA-Z0-9_$]+')

    WHITESPACE = purplex.TokenDef(r'[\s\n]+', ignore=True)


class JavaScriptParser(purplex.Parser):
    """Parser for a subset of JavaScript."""

    # pylint: disable=C0111,R0201,W0613,R0913
    LEXER = JavaScriptLexer
    START = 'e'
    PRECEDENCE = ()

    @purplex.attach('listitems : e')
    def listitems_1(self, child):
        return [child]

    @purplex.attach('listitems : e COMMA listitems')
    def listitems_2(self, child, comma, rest_of_list):
        return [child] + rest_of_list

    @purplex.attach('listitems : COMMA listitems')
    def listitems_3(self, comma, rest_of_list):
        return [None] + rest_of_list

    @purplex.attach('listitems : ')
    def listitems_4(self):
        return []

    @purplex.attach('e : LIST_START listitems LIST_END')
    def list(self, *children):
        return children[1]

    @purplex.attach('objectkey : e')
    @purplex.attach('objectkey : KEY')
    def objectkey(self, key):
        # TODO not everything can be a key
        return key

    @purplex.attach('objectitems : ')
    def objectitems_1(self):
        return {}

    @purplex.attach('objectitems : objectkey COLON e')
    def objectitems_2(self, key, colon, val):
        return {key: val}

    @purplex.attach('objectitems : objectkey COLON e COMMA objectitems')
    def objectitems_3(self, key, colon, val, comma, otheritems):
        d = dict(otheritems)
        d[key] = val
        return d

    @purplex.attach('e : OBJECT_START objectitems OBJECT_END')
    def object(self, start, objectitems, end):
        return objectitems

    @purplex.attach('e : INTEGER')
    def number(self, num):
        return int(num)

    @purplex.attach('e : FLOAT')
    def float_number(self, num):
        return float(num)

    @purplex.attach('e : NULL')
    def null(self, t):
        return None

    @purplex.attach('e : TRUE')
    def true(self, t):
        return True

    @purplex.attach('e : FALSE')
    def false(self, t):
        return False

    @purplex.attach('e : STRING')
    def string(self, s):
        return _unescape_string(s[1:-1])


# instantiate the parser at module-load time for better performance
_PARSER = JavaScriptParser()



from setuptools import setup
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


with open('README.rst') as f:
    readme = f.read()


setup(
    name='hangups',
    version='0.1.3',
    description=('the first third-party instant messaging client for Google '
                 'Hangouts'),
    long_description=readme,
    url='https://github.com/tdryer/hangups',
    author='Tom Dryer',
    author_email='tomdryer.com@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    packages=['hangups'],
    install_requires=[
        'purplex==0.2.4',
        'tornado==3.2.1',
        'requests==2.2.1',
        'urwid==1.2.1',
        'appdirs==1.3.0',
    ],
    tests_require=[
        'pytest',
    ],
    cmdclass={'test': PyTest},
    entry_points={
        'console_scripts': [
            'hangups=hangups.__main__:main',
        ],
    },
)
